#ifndef YERBACON_SEMANTICANALYSIS_HPP
#define YERBACON_SEMANTICANALYSIS_HPP

#include "parsing/ParseComponents.hpp"
#include "Tasking.hpp"

using namespace StandardComponents;

struct SemanticAnalyser final: public Tasking {
    typedef Yerbacon::Exception SemanticException;
    static void error(const string_view& text, unsigned long line = 0) {
        throw SemanticException(text, line);
    }
    const auto& analyseTree(const ParseTree& tree) {
        const auto& task_map = getTaskMapInstance();
        for (unsigned int i = 0; i < tree.size(); ++i) {
            try {
                task_map.at(tree[i]->getId())(tree, i);
            } catch (const out_of_range&) {}
        }
        return tree;
    }
private:
    unordered_task_map getTaskMapInstance() final {
        using namespace StandardComponents;
        return {
            make_nonlocal_task(Function,
                if (index > 0 and any_of(parsedTree.cbegin(), parsedTree.cbegin() + index, [&parseComponent](const component_ptr& pointer){
                    if (pointer->getId() != typeid(Reference)) {
                        string_view ptr_name;
                        try {
                            ptr_name = dynamic_cast<NamedIdentifier<true>&>(*pointer).name;
                        } catch (const bad_cast&) {
                            try {
                                ptr_name = dynamic_cast<NamedIdentifier<false>&>(*pointer).name;
                            } catch (const bad_cast&) {
                                return false;
                            }
                        };
                        if (ptr_name == parseComponent.name) {
                            return true;
                        }
                    }
                    return false;
                })) { error(parseComponent.name + " has already been defined previously", parseComponent.line); }
            ),
            share_task(Function, Class),
            share_task(Function, Define<true>),
            make_nonlocal_task(Define<false>,
                if (parsedTree.findReferenceByName<Define<true>>(parseComponent.name).has_value()) {
                    error(parseComponent.name + string(" cannot be redefined as it is final"), parseComponent.line);
                }
            )
        };
    };
};

#endif //YERBACON_SEMANTICANALYSIS_HPP
